"""Package initialization."""


__version__ = "2024.2.0"

__title__ = "edgetest"
__description__ = "Bleeding edge dependency testing"
__url__ = "https://github.com/capitalone/edgetest"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Akshay Gupta"
__email__ = "akshay.gupta2@capitalone.com"
